@@ -1,7 +1,7 @@
 /*
  *  OpenSlide, a library for reading whole slide image files
  *
- *  Copyright (c) 2007-2013 Carnegie Mellon University
+ *  Copyright (c) 2007-2015 Carnegie Mellon University
  *  Copyright (c) 2011 Google, Inc.
  *  All rights reserved.
  *
@@ -24,6 +24,7 @@
 
 #include "openslide-private.h"
 #include "openslide-decode-tiff.h"
+#include "openslide-decode-jpeg.h"
 
 #include <glib.h>
 #include <tiffio.h>
@@ -57,18 +58,16 @@ struct associated_image {
   tdir_t directory;
 };
 
-#define SET_DIR_OR_FAIL(tiff, i, err)					\
+#define SET_DIR_OR_FAIL(tiff, i)					\
   do {									\
-    if (!TIFFSetDirectory(tiff, i)) {					\
-      g_set_error(err, OPENSLIDE_ERROR, OPENSLIDE_ERROR_FAILED,		\
-                  "Cannot set TIFF directory %d", i);			\
+    if (!_openslide_tiff_set_dir(tiff, i, err)) {			\
       return false;							\
     }									\
   } while (0)
 
-#define GET_FIELD_OR_FAIL(tiff, tag, result, err)			\
+#define GET_FIELD_OR_FAIL(tiff, tag, type, result)			\
   do {									\
-    uint32 tmp;								\
+    type tmp;								\
     if (!TIFFGetField(tiff, tag, &tmp)) {				\
       g_set_error(err, OPENSLIDE_ERROR, OPENSLIDE_ERROR_FAILED,		\
                   "Cannot get required TIFF tag: %d", tag);		\
@@ -77,23 +76,56 @@ struct associated_image {
     result = tmp;							\
   } while (0)
 
+#undef TIFFSetDirectory
+bool _openslide_tiff_set_dir(TIFF *tiff,
+                             tdir_t dir,
+                             GError **err) {
+  if (dir == TIFFCurrentDirectory(tiff)) {
+    // avoid libtiff unnecessarily rereading directory contents
+    return true;
+  }
+  if (!TIFFSetDirectory(tiff, dir)) {
+    g_set_error(err, OPENSLIDE_ERROR, OPENSLIDE_ERROR_FAILED,
+                "Cannot set TIFF directory %d", dir);
+    return false;
+  }
+  return true;
+}
+#define TIFFSetDirectory _OPENSLIDE_POISON(_openslide_tiff_set_dir)
+
 bool _openslide_tiff_level_init(TIFF *tiff,
                                 tdir_t dir,
                                 struct _openslide_level *level,
                                 struct _openslide_tiff_level *tiffl,
                                 GError **err) {
   // set the directory
-  SET_DIR_OR_FAIL(tiff, dir, err);
+  SET_DIR_OR_FAIL(tiff, dir);
 
   // figure out tile size
   int64_t tw, th;
-  GET_FIELD_OR_FAIL(tiff, TIFFTAG_TILEWIDTH, tw, err);
-  GET_FIELD_OR_FAIL(tiff, TIFFTAG_TILELENGTH, th, err);
+  GET_FIELD_OR_FAIL(tiff, TIFFTAG_TILEWIDTH, uint32_t, tw);
+  GET_FIELD_OR_FAIL(tiff, TIFFTAG_TILELENGTH, uint32_t, th);
 
   // get image size
   int64_t iw, ih;
-  GET_FIELD_OR_FAIL(tiff, TIFFTAG_IMAGEWIDTH, iw, err);
-  GET_FIELD_OR_FAIL(tiff, TIFFTAG_IMAGELENGTH, ih, err);
+  GET_FIELD_OR_FAIL(tiff, TIFFTAG_IMAGEWIDTH, uint32_t, iw);
+  GET_FIELD_OR_FAIL(tiff, TIFFTAG_IMAGELENGTH, uint32_t, ih);
+
+  // decide whether we can bypass libtiff when reading tiles
+  uint16_t compression, planar_config, photometric;
+  uint16_t bits_per_sample, samples_per_pixel;
+  GET_FIELD_OR_FAIL(tiff, TIFFTAG_COMPRESSION, uint16_t, compression);
+  GET_FIELD_OR_FAIL(tiff, TIFFTAG_PLANARCONFIG, uint16_t, planar_config);
+  GET_FIELD_OR_FAIL(tiff, TIFFTAG_PHOTOMETRIC, uint16_t, photometric);
+  GET_FIELD_OR_FAIL(tiff, TIFFTAG_BITSPERSAMPLE, uint16_t, bits_per_sample);
+  GET_FIELD_OR_FAIL(tiff, TIFFTAG_SAMPLESPERPIXEL, uint16_t, samples_per_pixel);
+  bool read_direct =
+    compression == COMPRESSION_JPEG &&
+    planar_config == PLANARCONFIG_CONTIG &&
+    (photometric == PHOTOMETRIC_RGB || photometric == PHOTOMETRIC_YCBCR) &&
+    bits_per_sample == 8 &&
+    samples_per_pixel == 3;
+  //g_debug("directory %d, read_direct %d", dir, read_direct);
 
   // safe now, start writing
   if (level) {
@@ -114,6 +146,9 @@ bool _openslide_tiff_level_init(TIFF *tiff,
     // num tiles in each dimension
     tiffl->tiles_across = (iw / tw) + !!(iw % tw);   // integer ceiling
     tiffl->tiles_down = (ih / th) + !!(ih % th);
+
+    tiffl->tile_read_direct = read_direct;
+    tiffl->photometric = photometric;
   }
 
   return true;
@@ -174,18 +209,113 @@ static bool tiff_read_region(TIFF *tiff,
   return success;
 }
 
+static bool decode_jpeg(const void *buf, uint32_t buflen,
+                        const void *tables, uint32_t tables_len,  // optional
+                        J_COLOR_SPACE space,
+                        uint32_t *dest,
+                        int32_t w, int32_t h,
+                        GError **err) {
+  volatile bool result = false;
+  jmp_buf env;
+
+  struct jpeg_decompress_struct *cinfo;
+  struct _openslide_jpeg_decompress *dc =
+    _openslide_jpeg_decompress_create(&cinfo);
+
+  if (setjmp(env) == 0) {
+    _openslide_jpeg_decompress_init(dc, &env);
+
+    // load JPEG tables
+    if (tables) {
+      _openslide_jpeg_mem_src(cinfo, (void *) tables, tables_len);
+      if (jpeg_read_header(cinfo, false) != JPEG_HEADER_TABLES_ONLY) {
+        g_set_error(err, OPENSLIDE_ERROR, OPENSLIDE_ERROR_FAILED,
+                    "Couldn't load JPEG tables");
+        goto DONE;
+      }
+    }
+
+    // set up I/O
+    _openslide_jpeg_mem_src(cinfo, (void *) buf, buflen);
+
+    // read header
+    if (jpeg_read_header(cinfo, true) != JPEG_HEADER_OK) {
+      g_set_error(err, OPENSLIDE_ERROR, OPENSLIDE_ERROR_FAILED,
+                  "Couldn't read JPEG header");
+      goto DONE;
+    }
+
+    // set color space from TIFF photometric tag (for Aperio)
+    cinfo->jpeg_color_space = space;
+
+    // decompress
+    if (!_openslide_jpeg_decompress_run(dc, dest, false, w, h, err)) {
+      goto DONE;
+    }
+    result = true;
+  } else {
+    // setjmp has returned again
+    _openslide_jpeg_propagate_error(err, dc);
+  }
+
+DONE:
+  _openslide_jpeg_decompress_destroy(dc);
+
+  return result;
+}
+
 bool _openslide_tiff_read_tile(struct _openslide_tiff_level *tiffl,
                                TIFF *tiff,
                                uint32_t *dest,
                                int64_t tile_col, int64_t tile_row,
                                GError **err) {
   // set directory
-  SET_DIR_OR_FAIL(tiff, tiffl->dir, err);
+  SET_DIR_OR_FAIL(tiff, tiffl->dir);
 
-  // read tile
-  return tiff_read_region(tiff, dest,
-                          tile_col * tiffl->tile_w, tile_row * tiffl->tile_h,
-                          tiffl->tile_w, tiffl->tile_h, err);
+  if (tiffl->tile_read_direct) {
+    // Fast path: read raw data, decode through libjpeg
+    // Reading through tiff_read_region() reformats pixel data in three
+    // passes: libjpeg converts from planar to R G B, libtiff converts
+    // to BGRA, we convert to ARGB.  If we can bypass libtiff when
+    // decoding JPEG tiles, we can reduce this to one optimized pass in
+    // libjpeg-turbo.
+
+    // read tables
+    void *tables;
+    uint32_t tables_len;
+    if (!TIFFGetField(tiff, TIFFTAG_JPEGTABLES, &tables_len, &tables)) {
+      // no separate tables
+      tables = NULL;
+      tables_len = 0;
+    }
+
+    // read data
+    void *buf;
+    int32_t buflen;
+    if (!_openslide_tiff_read_tile_data(tiffl, tiff,
+                                        &buf, &buflen,
+                                        tile_col, tile_row,
+                                        err)) {
+      return false;
+    }
+
+    // decompress
+    bool ret = decode_jpeg(buf, buflen, tables, tables_len,
+                           tiffl->photometric == PHOTOMETRIC_YCBCR ? JCS_YCbCr : JCS_RGB,
+                           dest,
+                           tiffl->tile_w, tiffl->tile_h,
+                           err);
+    g_free(buf);
+    return ret;
+  } else {
+    // Fallback: read tile through libtiff
+    _openslide_performance_warn_once(&tiffl->warned_read_indirect,
+                                     "Using slow libtiff read path for "
+                                     "directory %d", tiffl->dir);
+    return tiff_read_region(tiff, dest,
+                            tile_col * tiffl->tile_w, tile_row * tiffl->tile_h,
+                            tiffl->tile_w, tiffl->tile_h, err);
+  }
 }
 
 bool _openslide_tiff_read_tile_data(struct _openslide_tiff_level *tiffl,
@@ -193,12 +323,8 @@ bool _openslide_tiff_read_tile_data(struct _openslide_tiff_level *tiffl,
                                     void **_buf, int32_t *_len,
                                     int64_t tile_col, int64_t tile_row,
                                     GError **err) {
-  // initialize out params
-  *_buf = NULL;
-  *_len = 0;
-
   // set directory
-  SET_DIR_OR_FAIL(tiff, tiffl->dir, err);
+  SET_DIR_OR_FAIL(tiff, tiffl->dir);
 
   // get tile number
   ttile_t tile_no = TIFFComputeTile(tiff,
@@ -217,12 +343,6 @@ bool _openslide_tiff_read_tile_data(struct _openslide_tiff_level *tiffl,
   }
   tsize_t tile_size = sizes[tile_no];
 
-  // a slide with zero-length tiles has been seen in the wild
-  if (!tile_size) {
-    //g_debug("no data for tile %d", tile_no);
-    return true;  // ok, haven't allocated anything yet
-  }
-
   // get raw tile
   tdata_t buf = g_malloc(tile_size);
   tsize_t size = TIFFReadRawTile(tiff, tile_no, buf, tile_size);
@@ -239,6 +359,40 @@ bool _openslide_tiff_read_tile_data(struct _openslide_tiff_level *tiffl,
   return true;
 }
 
+// sets out-argument to indicate whether the tile data is zero bytes long
+// returns false on error
+bool _openslide_tiff_check_missing_tile(struct _openslide_tiff_level *tiffl,
+                                        TIFF *tiff,
+                                        int64_t tile_col, int64_t tile_row,
+                                        bool *is_missing,
+                                        GError **err) {
+  // set directory
+  if (!_openslide_tiff_set_dir(tiff, tiffl->dir, err)) {
+    return false;
+  }
+
+  // get tile number
+  ttile_t tile_no = TIFFComputeTile(tiff,
+                                    tile_col * tiffl->tile_w,
+                                    tile_row * tiffl->tile_h,
+                                    0, 0);
+
+  //g_debug("_openslide_tiff_check_missing_tile: tile %d", tile_no);
+
+  // get tile size
+  toff_t *sizes;
+  if (!TIFFGetField(tiff, TIFFTAG_TILEBYTECOUNTS, &sizes)) {
+    g_set_error(err, OPENSLIDE_ERROR, OPENSLIDE_ERROR_FAILED,
+                "Cannot get tile size");
+    return false;
+  }
+  tsize_t tile_size = sizes[tile_no];
+
+  // return result
+  *is_missing = tile_size == 0;
+  return true;
+}
+
 static bool _get_associated_image_data(TIFF *tiff,
                                        struct associated_image *img,
                                        uint32_t *dest,
@@ -247,16 +401,15 @@ static bool _get_associated_image_data(TIFF *tiff,
 
   // g_debug("read TIFF associated image: %d", img->directory);
 
-  SET_DIR_OR_FAIL(tiff, img->directory, err);
+  SET_DIR_OR_FAIL(tiff, img->directory);
 
   // ensure dimensions have not changed
-  GET_FIELD_OR_FAIL(tiff, TIFFTAG_IMAGEWIDTH, width, err);
-  GET_FIELD_OR_FAIL(tiff, TIFFTAG_IMAGELENGTH, height, err);
+  GET_FIELD_OR_FAIL(tiff, TIFFTAG_IMAGEWIDTH, uint32_t, width);
+  GET_FIELD_OR_FAIL(tiff, TIFFTAG_IMAGELENGTH, uint32_t, height);
   if (img->base.w != width || img->base.h != height) {
     g_set_error(err, OPENSLIDE_ERROR, OPENSLIDE_ERROR_FAILED,
                 "Unexpected associated image size: "
-                "expected %"G_GINT64_FORMAT"x%"G_GINT64_FORMAT", "
-                "got %"G_GINT64_FORMAT"x%"G_GINT64_FORMAT,
+                "expected %"PRId64"x%"PRId64", got %"PRId64"x%"PRId64,
                 img->base.w, img->base.h, width, height);
     return false;
   }
@@ -296,16 +449,16 @@ static bool _add_associated_image(openslide_t *osr,
                                   TIFF *tiff,
                                   GError **err) {
   // set directory
-  SET_DIR_OR_FAIL(tiff, dir, err);
+  SET_DIR_OR_FAIL(tiff, dir);
 
   // get the dimensions
   int64_t w, h;
-  GET_FIELD_OR_FAIL(tiff, TIFFTAG_IMAGEWIDTH, w, err);
-  GET_FIELD_OR_FAIL(tiff, TIFFTAG_IMAGELENGTH, h, err);
+  GET_FIELD_OR_FAIL(tiff, TIFFTAG_IMAGEWIDTH, uint32_t, w);
+  GET_FIELD_OR_FAIL(tiff, TIFFTAG_IMAGELENGTH, uint32_t, h);
 
   // check compression
   uint16_t compression;
-  GET_FIELD_OR_FAIL(tiff, TIFFTAG_COMPRESSION, compression, err);
+  GET_FIELD_OR_FAIL(tiff, TIFFTAG_COMPRESSION, uint16_t, compression);
   if (!TIFFIsCODECConfigured(compression)) {
     g_set_error(err, OPENSLIDE_ERROR, OPENSLIDE_ERROR_FAILED,
                 "Unsupported TIFF compression: %u", compression);
@@ -337,6 +490,9 @@ bool _openslide_tiff_add_associated_image(openslide_t *osr,
     ret = _add_associated_image(osr, name, tc, dir, tiff, err);
   }
   _openslide_tiffcache_put(tc, tiff);
+
+  // safe even if successful
+  g_prefix_error(err, "Can't read %s associated image: ", name);
   return ret;
 }
 
@@ -398,6 +554,7 @@ static toff_t tiff_do_size(thandle_t th) {
   return hdl->size;
 }
 
+#undef TIFFClientOpen
 static TIFF *tiff_open(struct _openslide_tiffcache *tc, GError **err) {
   // open
   FILE *f = _openslide_fopen(tc->filename, "rb", err);
@@ -478,6 +635,7 @@ NOT_TIFF:
               "Not a TIFF file: %s", tc->filename);
   return NULL;
 }
+#define TIFFClientOpen _OPENSLIDE_POISON(_openslide_tiffcache_get)
 
 struct _openslide_tiffcache *_openslide_tiffcache_create(const char *filename) {
   struct _openslide_tiffcache *tc = g_slice_new0(struct _openslide_tiffcache);
