@@ -1,7 +1,7 @@
 /*
  *  OpenSlide, a library for reading whole slide image files
  *
- *  Copyright (c) 2007-2013 Carnegie Mellon University
+ *  Copyright (c) 2007-2015 Carnegie Mellon University
  *  Copyright (c) 2011 Google, Inc.
  *  All rights reserved.
  *
@@ -27,7 +27,6 @@
 #include <stdio.h>
 #include <stdlib.h>
 #include <string.h>
-#include <stdint.h>
 #include <math.h>
 #include <glib.h>
 
@@ -44,14 +43,22 @@
 
 #define NO_OFFSET UINT64_MAX
 
+#define NDPI_TAG 65420
+
 
 struct _openslide_tifflike {
   char *filename;
   bool big_endian;
+  bool ndpi;
   GPtrArray *directories;
   GMutex *value_lock;
 };
 
+struct tiff_directory {
+  GHashTable *items;
+  uint64_t offset;  // only for NDPI fixups
+};
+
 struct tiff_item {
   uint16_t type;
   int64_t count;
@@ -168,6 +175,18 @@ static uint32_t get_value_size(uint16_t type, uint64_t *count) {
   }
 }
 
+// Re-add implied high-order bits to a 32-bit offset.
+// Heuristic: maximize high-order bits while keeping the offset below diroff.
+static uint64_t fix_offset_ndpi(uint64_t diroff, uint64_t offset) {
+  uint64_t result = (diroff & ~(uint64_t) UINT32_MAX) | (offset & UINT32_MAX);
+  if (result >= diroff) {
+    // ensure result doesn't wrap around
+    result = MIN(result - UINT32_MAX - 1, result);
+  }
+  //g_debug("diroff %"PRIx64": %"PRIx64" -> %"PRIx64, diroff, offset, result);
+  return result;
+}
+
 #define ALLOC_VALUES_OR_FAIL(OUT, TYPE, COUNT) do {			\
     OUT = g_try_new(TYPE, COUNT);					\
     if (!OUT) {								\
@@ -338,7 +357,7 @@ static bool populate_item(struct _openslide_tifflike *tl,
     goto FAIL;
   }
 
-  //g_debug("reading tiff value: len: %"G_GINT64_FORMAT", offset %"G_GINT64_FORMAT, len, item->offset);
+  //g_debug("reading tiff value: len: %"PRId64", offset %"PRIu64, len, item->offset);
   if (fseeko(f, item->offset, SEEK_SET)) {
     _openslide_io_error(err, "Couldn't seek to read TIFF value");
     goto FAIL;
@@ -365,6 +384,14 @@ FAIL:
   return success;
 }
 
+static void tiff_directory_destroy(struct tiff_directory *d) {
+  if (d == NULL) {
+    return;
+  }
+  g_hash_table_unref(d->items);
+  g_slice_free(struct tiff_directory, d);
+}
+
 static void tiff_item_destroy(gpointer data) {
   struct tiff_item *item = data;
 
@@ -375,17 +402,19 @@ static void tiff_item_destroy(gpointer data) {
   g_slice_free(struct tiff_item, item);
 }
 
-static GHashTable *read_directory(FILE *f, int64_t *diroff,
-				  GHashTable *loop_detector,
-				  bool bigtiff,
-				  bool big_endian,
-				  GError **err) {
+static struct tiff_directory *read_directory(FILE *f, int64_t *diroff,
+                                             struct tiff_directory *first_dir,
+                                             GHashTable *loop_detector,
+                                             bool bigtiff,
+                                             bool ndpi,
+                                             bool big_endian,
+                                             GError **err) {
   int64_t off = *diroff;
   *diroff = 0;
-  GHashTable *result = NULL;
+  struct tiff_directory *d = NULL;
   bool ok = true;
 
-  //  g_debug("diroff: %" PRId64, off);
+  //  g_debug("diroff: %"PRId64, off);
 
   if (off <= 0) {
     g_set_error(err, OPENSLIDE_ERROR, OPENSLIDE_ERROR_FAILED,
@@ -418,12 +447,14 @@ static GHashTable *read_directory(FILE *f, int64_t *diroff,
     goto FAIL;
   }
 
-  //  g_debug("dircount: %"G_GUINT64_FORMAT, dircount);
+  //  g_debug("dircount: %"PRIu64, dircount);
 
 
-  // initial checks passed, initialize the hashtable
-  result = g_hash_table_new_full(g_direct_hash, g_direct_equal,
-                                 NULL, tiff_item_destroy);
+  // initial checks passed, initialize the directory
+  d = g_slice_new0(struct tiff_directory);
+  d->items = g_hash_table_new_full(g_direct_hash, g_direct_equal,
+                                   NULL, tiff_item_destroy);
+  d->offset = off;
 
   // read all directory entries
   for (uint64_t n = 0; n < dircount; n++) {
@@ -437,13 +468,13 @@ static GHashTable *read_directory(FILE *f, int64_t *diroff,
       goto FAIL;
     }
 
-    //    g_debug(" tag: %d, type: %d, count: %" PRId64, tag, type, count);
+    //    g_debug(" tag: %d, type: %d, count: %"PRId64, tag, type, count);
 
     // allocate the item
     struct tiff_item *item = g_slice_new0(struct tiff_item);
     item->type = type;
     item->count = count;
-    g_hash_table_insert(result, GINT_TO_POINTER(tag), item);
+    g_hash_table_insert(d->items, GINT_TO_POINTER(tag), item);
 
     // compute value size
     uint32_t value_size = get_value_size(type, &count);
@@ -487,11 +518,25 @@ static GHashTable *read_directory(FILE *f, int64_t *diroff,
         fix_byte_order(&off32, sizeof(off32), 1, big_endian);
         item->offset = off32;
       }
+
+      if (ndpi) {
+        // heuristically set high-order bits of offset
+        // if this tag has the same offset in the first IFD, reuse that value
+        struct tiff_item *first_dir_item = NULL;
+        if (first_dir) {
+          first_dir_item = g_hash_table_lookup(first_dir->items,
+                                               GINT_TO_POINTER(tag));
+        }
+        if (!first_dir_item || first_dir_item->offset != item->offset) {
+          item->offset = fix_offset_ndpi(off, item->offset);
+        }
+      }
     }
   }
 
   // read the next dir offset
-  int64_t nextdiroff = read_uint(f, bigtiff ? 8 : 4, big_endian, &ok);
+  int64_t nextdiroff = read_uint(f, (bigtiff || ndpi) ? 8 : 4,
+                                 big_endian, &ok);
   if (!ok) {
     g_set_error(err, OPENSLIDE_ERROR, OPENSLIDE_ERROR_FAILED,
                 "Cannot read next directory offset");
@@ -500,13 +545,11 @@ static GHashTable *read_directory(FILE *f, int64_t *diroff,
   *diroff = nextdiroff;
 
   // success
-  return result;
+  return d;
 
 
- FAIL:
-  if (result != NULL) {
-    g_hash_table_unref(result);
-  }
+FAIL:
+  tiff_directory_destroy(d);
   return NULL;
 }
 
@@ -523,7 +566,6 @@ struct _openslide_tifflike *_openslide_tifflike_create(const char *filename,
 
   // read and check magic
   uint16_t magic;
-  fseeko(f, 0, SEEK_SET);
   if (fread(&magic, sizeof magic, 1, f) != 1) {
     g_set_error(err, OPENSLIDE_ERROR, OPENSLIDE_ERROR_FAILED,
                 "Can't read TIFF magic number");
@@ -548,7 +590,8 @@ struct _openslide_tifflike *_openslide_tifflike_create(const char *filename,
     offset_size = read_uint(f, 2, big_endian, &ok);
     pad = read_uint(f, 2, big_endian, &ok);
   }
-  int64_t diroff = read_uint(f, bigtiff ? 8 : 4, big_endian, &ok);
+  // for classic TIFF, will mask off the high bytes after NDPI detection
+  int64_t diroff = read_uint(f, 8, big_endian, &ok);
 
   if (!ok) {
     g_set_error(err, OPENSLIDE_ERROR, OPENSLIDE_ERROR_FAILED,
@@ -578,24 +621,69 @@ struct _openslide_tifflike *_openslide_tifflike_create(const char *filename,
   tl->directories = g_ptr_array_new();
   tl->value_lock = g_mutex_new();
 
-  // initialize loop detector
+  // initialize directory reading
   loop_detector = g_hash_table_new_full(_openslide_int64_hash,
                                         _openslide_int64_equal,
                                         _openslide_int64_free,
                                         NULL);
+  struct tiff_directory *first_dir = NULL;
+
+  // NDPI needs special quirks, since it is classic TIFF pretending to be
+  // BigTIFF.  Enable NDPI mode if this is classic TIFF but the offset to
+  // the first directory -- when treated as a 64-bit value -- points to a
+  // valid directory containing the NDPI_TAG.
+  if (!bigtiff && diroff != 0) {
+    int64_t trial_diroff = diroff;
+    struct tiff_directory *d = read_directory(f, &trial_diroff,
+                                              NULL,
+                                              loop_detector,
+                                              bigtiff, true, big_endian,
+                                              NULL);
+    if (d) {
+      struct tiff_item *item =
+        g_hash_table_lookup(d->items, GINT_TO_POINTER(NDPI_TAG));
+      if (item && item->count) {
+        // NDPI
+        //g_debug("NDPI detected");
+        tl->ndpi = true;
+        // save the parsed directory rather than reparsing it below
+        g_ptr_array_add(tl->directories, d);
+        first_dir = d;
+        diroff = trial_diroff;
+      } else {
+        // correctly parsed the directory in NDPI mode, but didn't find
+        // NDPI_TAG
+        tiff_directory_destroy(d);
+      }
+    }
+    if (!tl->ndpi) {
+      // This is classic TIFF, so diroff is 32 bits.  Mask off the high bits
+      // and reset.
+      //g_debug("not NDPI");
+      diroff &= 0xffffffff;
+      g_hash_table_remove_all(loop_detector);
+    }
+  }
+
   // read all the directories
   while (diroff != 0) {
     // read a directory
-    GHashTable *ht = read_directory(f, &diroff, loop_detector, bigtiff,
-                                    big_endian, err);
+    struct tiff_directory *d = read_directory(f, &diroff,
+                                              first_dir,
+                                              loop_detector,
+                                              bigtiff, tl->ndpi, big_endian,
+                                              err);
 
     // was the directory successfully read?
-    if (ht == NULL) {
+    if (d == NULL) {
       goto FAIL;
     }
 
-    // add result to array
-    g_ptr_array_add(tl->directories, ht);
+    // store result
+    g_ptr_array_add(tl->directories, d);
+    if (!first_dir) {
+      first_dir = d;
+    }
   }
 
   // ensure there are directories
@@ -627,7 +715,7 @@ void _openslide_tifflike_destroy(struct _openslide_tifflike *tl) {
   }
   g_mutex_lock(tl->value_lock);
   for (uint32_t n = 0; n < tl->directories->len; n++) {
-    g_hash_table_unref(tl->directories->pdata[n]);
+    tiff_directory_destroy(tl->directories->pdata[n]);
   }
   g_mutex_unlock(tl->value_lock);
   g_ptr_array_free(tl->directories, true);
@@ -641,8 +729,8 @@ static struct tiff_item *get_item(struct _openslide_tifflike *tl,
   if (dir < 0 || dir >= tl->directories->len) {
     return NULL;
   }
-  return g_hash_table_lookup(tl->directories->pdata[dir],
-                             GINT_TO_POINTER(tag));
+  struct tiff_directory *d = tl->directories->pdata[dir];
+  return g_hash_table_lookup(d->items, GINT_TO_POINTER(tag));
 }
 
 static void print_tag(struct _openslide_tifflike *tl,
@@ -650,7 +738,7 @@ static void print_tag(struct _openslide_tifflike *tl,
   struct tiff_item *item = get_item(tl, dir, tag);
   g_assert(item != NULL);
 
-  printf(" %d: type: %d, count: %" G_GINT64_FORMAT "\n ", tag, item->type, item->count);
+  printf(" %d: type: %d, count: %"PRId64"\n ", tag, item->type, item->count);
 
   switch (item->type) {
   case TIFF_ASCII: {
@@ -674,7 +762,7 @@ static void print_tag(struct _openslide_tifflike *tl,
   case TIFF_LONG8: {
     const uint64_t *uints = _openslide_tifflike_get_uints(tl, dir, tag, NULL);
     for (int64_t i = 0; i < item->count; i++) {
-      printf(" %" G_GUINT64_FORMAT, uints[i]);
+      printf(" %"PRIu64, uints[i]);
     }
     break;
   }
@@ -683,7 +771,7 @@ static void print_tag(struct _openslide_tifflike *tl,
   case TIFF_IFD8: {
     const uint64_t *uints = _openslide_tifflike_get_uints(tl, dir, tag, NULL);
     for (int64_t i = 0; i < item->count; i++) {
-      printf(" %.16" G_GINT64_MODIFIER "x", uints[i]);
+      printf(" %.16"PRIx64, uints[i]);
     }
     break;
   }
@@ -694,7 +782,7 @@ static void print_tag(struct _openslide_tifflike *tl,
   case TIFF_SLONG8: {
     const int64_t *sints = _openslide_tifflike_get_sints(tl, dir, tag, NULL);
     for (int64_t i = 0; i < item->count; i++) {
-      printf(" %" G_GINT64_FORMAT, sints[i]);
+      printf(" %"PRId64, sints[i]);
     }
     break;
   }
@@ -731,7 +819,8 @@ static int tag_compare(gconstpointer a, gconstpointer b) {
 
 static void print_directory(struct _openslide_tifflike *tl,
                             int64_t dir) {
-  GList *keys = g_hash_table_get_keys(tl->directories->pdata[dir]);
+  struct tiff_directory *d = tl->directories->pdata[dir];
+  GList *keys = g_hash_table_get_keys(d->items);
   keys = g_list_sort(keys, tag_compare);
   for (GList *el = keys; el; el = el->next) {
     print_tag(tl, dir, GPOINTER_TO_INT(el->data));
@@ -767,8 +856,7 @@ static struct tiff_item *get_and_check_item(struct _openslide_tifflike *tl,
   struct tiff_item *item = get_item(tl, dir, tag);
   if (item == NULL || item->count == 0) {
     g_set_error(err, OPENSLIDE_ERROR, OPENSLIDE_ERROR_NO_VALUE,
-                "No such value: directory %"G_GINT64_FORMAT", tag %d",
-                dir, tag);
+                "No such value: directory %"PRId64", tag %d", dir, tag);
     return NULL;
   }
   return item;
@@ -783,7 +871,7 @@ uint64_t _openslide_tifflike_get_uint(struct _openslide_tifflike *tl,
   }
   if (!item->uints) {
     g_set_error(err, OPENSLIDE_ERROR, OPENSLIDE_ERROR_FAILED,
-                "Unexpected value type: directory %"G_GINT64_FORMAT", "
+                "Unexpected value type: directory %"PRId64", "
                 "tag %d, type %d", dir, tag, item->type);
     return 0;
   }
@@ -799,7 +887,7 @@ int64_t _openslide_tifflike_get_sint(struct _openslide_tifflike *tl,
   }
   if (!item->sints) {
     g_set_error(err, OPENSLIDE_ERROR, OPENSLIDE_ERROR_FAILED,
-                "Unexpected value type: directory %"G_GINT64_FORMAT", "
+                "Unexpected value type: directory %"PRId64", "
                 "tag %d, type %d", dir, tag, item->type);
     return 0;
   }
@@ -815,7 +903,7 @@ double _openslide_tifflike_get_float(struct _openslide_tifflike *tl,
   }
   if (!item->floats) {
     g_set_error(err, OPENSLIDE_ERROR, OPENSLIDE_ERROR_FAILED,
-                "Unexpected value type: directory %"G_GINT64_FORMAT", "
+                "Unexpected value type: directory %"PRId64", "
                 "tag %d, type %d", dir, tag, item->type);
     return NAN;
   }
@@ -831,7 +919,7 @@ const uint64_t *_openslide_tifflike_get_uints(struct _openslide_tifflike *tl,
   }
   if (!item->uints) {
     g_set_error(err, OPENSLIDE_ERROR, OPENSLIDE_ERROR_FAILED,
-                "Unexpected value type: directory %"G_GINT64_FORMAT", "
+                "Unexpected value type: directory %"PRId64", "
                 "tag %d, type %d", dir, tag, item->type);
     return NULL;
   }
@@ -847,7 +935,7 @@ const int64_t *_openslide_tifflike_get_sints(struct _openslide_tifflike *tl,
   }
   if (!item->sints) {
     g_set_error(err, OPENSLIDE_ERROR, OPENSLIDE_ERROR_FAILED,
-                "Unexpected value type: directory %"G_GINT64_FORMAT", "
+                "Unexpected value type: directory %"PRId64", "
                 "tag %d, type %d", dir, tag, item->type);
     return NULL;
   }
@@ -863,7 +951,7 @@ const double *_openslide_tifflike_get_floats(struct _openslide_tifflike *tl,
   }
   if (!item->floats) {
     g_set_error(err, OPENSLIDE_ERROR, OPENSLIDE_ERROR_FAILED,
-                "Unexpected value type: directory %"G_GINT64_FORMAT", "
+                "Unexpected value type: directory %"PRId64", "
                 "tag %d, type %d", dir, tag, item->type);
     return NULL;
   }
@@ -879,7 +967,7 @@ const void *_openslide_tifflike_get_buffer(struct _openslide_tifflike *tl,
   }
   if (!item->buffer) {
     g_set_error(err, OPENSLIDE_ERROR, OPENSLIDE_ERROR_FAILED,
-                "Unexpected value type: directory %"G_GINT64_FORMAT", "
+                "Unexpected value type: directory %"PRId64", "
                 "tag %d, type %d", dir, tag, item->type);
     return NULL;
   }
@@ -892,6 +980,16 @@ bool _openslide_tifflike_is_tiled(struct _openslide_tifflike *tl,
          _openslide_tifflike_get_value_count(tl, dir, TIFFTAG_TILELENGTH);
 }
 
+uint64_t _openslide_tifflike_uint_fix_offset_ndpi(struct _openslide_tifflike *tl,
+                                                  int64_t dir, uint64_t offset) {
+  g_assert(dir >= 0 && dir < tl->directories->len);
+  if (!tl->ndpi) {
+    return offset;
+  }
+  struct tiff_directory *d = tl->directories->pdata[dir];
+  return fix_offset_ndpi(d->offset, offset);
+}
+
 static const char *store_string_property(struct _openslide_tifflike *tl,
                                          int64_t dir,
                                          openslide_t *osr,
