@@ -22,6 +22,7 @@
 #include <config.h>
 
 #include "openslide-private.h"
+#include "openslide-error.h"
 
 #include <glib.h>
 #include <stdarg.h>
@@ -33,35 +34,14 @@ const char *openslide_get_error(openslide_t *osr) {
 }
 
 // private error functions
-bool _openslide_set_error(openslide_t *osr, const char *format, ...) {
-  g_assert(format != NULL);
-
-  va_list args;
-
-  // format it for us
-  va_start(args, format);
-  char *newmsg = g_strdup_vprintf(format, args);
-  va_end(args);
-
-  if (!g_atomic_pointer_compare_and_exchange(&osr->error, NULL, newmsg)) {
+void _openslide_propagate_error(openslide_t *osr, GError *err) {
+  g_return_if_fail(err);
+  gchar *msg = g_strdup(err->message);
+  if (!g_atomic_pointer_compare_and_exchange(&osr->error, NULL, msg)) {
     // didn't replace the error, free it
-    g_free(newmsg);
-    return false;
-  } else {
-    // error was set
-    return true;
+    g_free(msg);
   }
-}
-
-bool _openslide_check_cairo_status_possibly_set_error(openslide_t *osr, cairo_t *cr) {
-  cairo_status_t status = cairo_status(cr);
-  if (!status) {
-    return false;
-  }
-
-  // cairo has error, let's set our error from it
-  return _openslide_set_error(osr, "cairo error: %s",
-			      cairo_status_to_string(status));
+  g_error_free(err);
 }
 
 // internal error propagation
@@ -81,7 +61,14 @@ void _openslide_io_error(GError **err, const char *fmt, ...) {
   va_end(ap);
 }
 
-void _openslide_set_error_from_gerror(openslide_t *osr, GError *err) {
-  g_return_if_fail(err);
-  _openslide_set_error(osr, "%s", err->message);
+bool _openslide_check_cairo_status(cairo_t *cr, GError **err) {
+  cairo_status_t status = cairo_status(cr);
+  if (!status) {
+    return true;
+  }
+
+  // cairo has error; set GError from it
+  g_set_error(err, OPENSLIDE_ERROR, OPENSLIDE_ERROR_CAIRO_ERROR,
+              "cairo error: %s", cairo_status_to_string(status));
+  return false;
 }
