@@ -80,10 +80,6 @@ static const char VALUE_SCAN_DATA_LAYER_LABEL[] = "ScanDataLayer_SlideBarcode";
 static const char VALUE_SCAN_DATA_LAYER_THUMBNAIL[] = "ScanDataLayer_SlidePreview";
 static const char VALUE_SLIDE_ZOOM_LEVEL[] = "Slide zoom level";
 
-static const char GROUP_NONHIERLAYER_d_SECTION[] = "NONHIERLAYER_%d_SECTION";
-static const char KEY_VIMSLIDE_POSITION_DATA_FORMAT_VERSION[] =
-  "VIMSLIDE_POSITION_DATA_FORMAT_VERSION";
-static const int VALUE_VIMSLIDE_POSITION_DATA_FORMAT_VERSION = 257;
 static const int SLIDE_POSITION_RECORD_SIZE = 9;
 
 static const char GROUP_DATAFILE[] = "DATAFILE";
@@ -179,7 +175,8 @@ struct slide_zoom_level_params {
 
 struct image {
   int32_t fileno;
-  int64_t start_in_file;
+  int32_t start_in_file;
+  int32_t length;
   int32_t imageno;   // used only for cache lookup
   int refcount;
 };
@@ -244,8 +241,9 @@ static uint32_t *read_image(openslide_t *osr,
                                  err);
     break;
   case FORMAT_BMP:
-    result = _openslide_gdkpixbuf_read(data->datafile_paths[image->fileno],
-                                       image->start_in_file,
+    result = _openslide_gdkpixbuf_read("bmp",
+                                       data->datafile_paths[image->fileno],
+                                       image->start_in_file, image->length,
                                        dest, w, h,
                                        err);
     break;
@@ -840,6 +838,7 @@ static bool process_hier_data_pages_from_indexfile(FILE *f,
 	struct image *image = g_slice_new0(struct image);
 	image->fileno = fileno;
 	image->start_in_file = offset;
+	image->length = length;
 	image->imageno = image_number++;
 	image->refcount = 1;
 
@@ -947,7 +946,7 @@ static void *inflate_buffer(const void *src,
 ZLIB_ERROR:
   if (error_code == Z_STREAM_END) {
     g_set_error(err, OPENSLIDE_ERROR, OPENSLIDE_ERROR_FAILED,
-                "Short read while decompressing: %lu/%"G_GINT64_FORMAT,
+                "Short read while decompressing: %lu/%"PRId64,
                 strm.total_out, dst_len);
   } else if (strm.msg) {
     g_set_error(err, OPENSLIDE_ERROR, OPENSLIDE_ERROR_FAILED,
@@ -1062,23 +1061,20 @@ static bool add_associated_image(openslide_t *osr,
   char *path;
   int64_t size;
   int64_t offset;
-  bool result = false;
 
   if (recordno == -1) {
     // no such image
     return true;
   }
 
-  if (read_nonhier_record(indexfile, nonhier_root,
-                          datafile_count, datafile_paths, recordno,
-                          &path, &size, &offset, err)) {
-    result = _openslide_jpeg_add_associated_image(osr, name, path, offset, err);
+  if (!read_nonhier_record(indexfile, nonhier_root,
+                           datafile_count, datafile_paths, recordno,
+                           &path, &size, &offset, err)) {
+    g_prefix_error(err, "Cannot read %s associated image: ", name);
+    return false;
   }
 
-  if (!result) {
-    g_prefix_error(err, "Cannot read %s associated image: ", name);
-  }
-  return result;
+  return _openslide_jpeg_add_associated_image(osr, name, path, offset, err);
 }
 
 
@@ -1768,7 +1764,7 @@ static bool mirax_open(openslide_t *osr, const char *filename,
     g_debug("  overlap_y: %g", hs->overlap_y);
     g_debug("  mpp_x: %g", hs->mpp_x);
     g_debug("  mpp_y: %g", hs->mpp_y);
-    g_debug("  fill_rgb: %" G_GUINT32_FORMAT, hs->fill_rgb);
+    g_debug("  fill_rgb: %"PRIu32, hs->fill_rgb);
     g_debug("  image_w: %d", hs->image_w);
     g_debug("  image_h: %d", hs->image_h);
   }
@@ -1936,7 +1932,7 @@ static bool mirax_open(openslide_t *osr, const char *filename,
                                              lp->tile_advance_y,
                                              read_tile, tile_free);
 
-    //g_debug("level %d tile advance %.10g %.10g, dim %" G_GINT64_FORMAT " %" G_GINT64_FORMAT ", image size %d %d, tile %g %g, image_concat %d, tile_count_divisor %d, positions_per_tile %d", i, lp->tile_advance_x, lp->tile_advance_y, l->base.w, l->base.h, l->image_width, l->image_height, l->tile_w, l->tile_h, lp->image_concat, lp->tile_count_divisor, lp->positions_per_tile);
+    //g_debug("level %d tile advance %.10g %.10g, dim %"PRId64" %"PRId64", image size %d %d, tile %g %g, image_concat %d, tile_count_divisor %d, positions_per_tile %d", i, lp->tile_advance_x, lp->tile_advance_y, l->base.w, l->base.h, l->image_width, l->image_height, l->tile_w, l->tile_h, lp->image_concat, lp->tile_count_divisor, lp->positions_per_tile);
   }
 
   // load the position map and build up the tiles
@@ -1982,7 +1978,7 @@ static bool mirax_open(openslide_t *osr, const char *filename,
   for (int i = 0; i < zoom_levels; i++) {
     struct level *l = levels[i];
     g_debug("level %d", i);
-    g_debug(" size %" G_GINT64_FORMAT " %" G_GINT64_FORMAT, l->base.w, l->base.h);
+    g_debug(" size %"PRId64" %"PRId64, l->base.w, l->base.h);
     g_debug(" image size %d %d", l->image_width, l->image_height);
     g_debug(" tile advance %g %g", lp->tile_advance_x, lp->tile_advance_y);
   }
