@@ -1,7 +1,7 @@
 /*
  *  OpenSlide, a library for reading whole slide image files
  *
- *  Copyright (c) 2007-2014 Carnegie Mellon University
+ *  Copyright (c) 2007-2015 Carnegie Mellon University
  *  Copyright (c) 2011 Google, Inc.
  *  All rights reserved.
  *
@@ -21,7 +21,7 @@
  */
 
 /*
- * Ventana (bif) support
+ * Ventana (bif/tif) support
  *
  * quickhash comes from _openslide_tifflike_init_properties_and_hash
  *
@@ -41,23 +41,24 @@
 #include <libxml/tree.h>
 #include <libxml/xpath.h>
 
-static const char LEVEL0_IMAGEDESCRIPTION_MAGIC[] = "level=0";
-static const char LEVEL0_XML_MAGIC[] = "iScan";
-static const char LEVEL0_XML_ROOT[] = "EncodeInfo";
-
 static const char LEVEL_DESCRIPTION_TOKEN[] = "level=";
 static const char MACRO_DESCRIPTION[] = "Label Image";
+static const char MACRO_DESCRIPTION2[] = "Label_Image";
 static const char THUMBNAIL_DESCRIPTION[] = "Thumbnail";
 
 static const char LEVEL_KEY[] = "level";
 static const char MAGNIFICATION_KEY[] = "mag";
 
-static const char ATTR_Z_LAYERS[] = "Z-layers";
+static const char INITIAL_XML_ISCAN[] = "iScan";
+static const char INITIAL_XML_ALT_ROOT[] = "Metadata";
+
 static const char ATTR_AOI_SCANNED[] = "AOIScanned";
 static const char ATTR_WIDTH[] = "Width";
 static const char ATTR_HEIGHT[] = "Height";
 static const char ATTR_NUM_ROWS[] = "NumRows";
 static const char ATTR_NUM_COLS[] = "NumCols";
+static const char ATTR_POS_X[] = "Pos-X";
+static const char ATTR_POS_Y[] = "Pos-Y";
 static const char ATTR_ORIGIN_X[] = "OriginX";
 static const char ATTR_ORIGIN_Y[] = "OriginY";
 static const char ATTR_CONFIDENCE[] = "Confidence";
@@ -97,10 +98,11 @@ struct level {
   struct _openslide_level base;
   struct _openslide_tiff_level tiffl;
   struct _openslide_grid *grid;
+  int64_t subtiles_per_tile;
 };
 
-// structs used during open
-struct slide_info {
+// structs used during BIF open
+struct bif {
   struct area **areas;
   int32_t num_areas;
 
@@ -109,6 +111,8 @@ struct slide_info {
 };
 
 struct area {
+  int64_t x;
+  int64_t y;
   int64_t start_col;
   int64_t start_row;
   int64_t tiles_across;
@@ -145,26 +149,24 @@ static bool read_subtile(openslide_t *osr,
                          cairo_t *cr,
                          struct _openslide_level *level,
                          int64_t subtile_col, int64_t subtile_row,
-                         void *subtile G_GNUC_UNUSED,
                          void *arg,
                          GError **err) {
   struct level *l = (struct level *) level;
   struct _openslide_tiff_level *tiffl = &l->tiffl;
   TIFF *tiff = arg;
-  const int64_t subtiles_per_tile = l->base.downsample;
   bool success = true;
 
   // tile size and coordinates
-  int64_t tile_col = subtile_col / subtiles_per_tile;
-  int64_t tile_row = subtile_row / subtiles_per_tile;
+  int64_t tile_col = subtile_col / l->subtiles_per_tile;
+  int64_t tile_row = subtile_row / l->subtiles_per_tile;
   int64_t tw = tiffl->tile_w;
   int64_t th = tiffl->tile_h;
 
   // subtile offset and size
-  double subtile_w = (double) tw / subtiles_per_tile;
-  double subtile_h = (double) th / subtiles_per_tile;
-  double subtile_x = subtile_col % subtiles_per_tile * subtile_w;
-  double subtile_y = subtile_row % subtiles_per_tile * subtile_h;
+  double subtile_w = (double) tw / l->subtiles_per_tile;
+  double subtile_h = (double) th / l->subtiles_per_tile;
+  double subtile_x = subtile_col % l->subtiles_per_tile * subtile_w;
+  double subtile_y = subtile_row % l->subtiles_per_tile * subtile_h;
 
   // get tile data, possibly from cache
   struct _openslide_cache_entry *cache_entry;
@@ -202,7 +204,7 @@ static bool read_subtile(openslide_t *osr,
 
   // if we are drawing a subtile, we must do an additional copy,
   // because cairo lacks source clipping
-  if (subtiles_per_tile > 1) {
+  if (l->subtiles_per_tile > 1) {
     cairo_surface_t *surface2 = cairo_image_surface_create(CAIRO_FORMAT_ARGB32,
                                                            ceil(subtile_w),
                                                            ceil(subtile_h));
@@ -231,6 +233,20 @@ static bool read_subtile(openslide_t *osr,
   return success;
 }
 
+// read_subtile wrapper for BIF that drops the extra argument passed by
+// the tilemap grid
+static bool read_subtile_tilemap(openslide_t *osr,
+                                 cairo_t *cr,
+                                 struct _openslide_level *level,
+                                 int64_t subtile_col, int64_t subtile_row,
+                                 void *subtile G_GNUC_UNUSED,
+                                 void *arg,
+                                 GError **err) {
+  return read_subtile(osr, cr, level,
+                      subtile_col, subtile_row,
+                      arg, err);
+}
+
 static bool paint_region(openslide_t *osr, cairo_t *cr,
                          int64_t x, int64_t y,
                          struct _openslide_level *level,
@@ -238,23 +254,17 @@ static bool paint_region(openslide_t *osr, cairo_t *cr,
                          GError **err) {
   struct ventana_ops_data *data = osr->data;
   struct level *l = (struct level *) level;
-  bool success = false;
 
   TIFF *tiff = _openslide_tiffcache_get(data->tc, err);
   if (tiff == NULL) {
     return false;
   }
 
-  if (TIFFSetDirectory(tiff, l->tiffl.dir)) {
-    success = _openslide_grid_paint_region(l->grid, cr, tiff,
-                                           x / l->base.downsample,
-                                           y / l->base.downsample,
-                                           level, w, h,
-                                           err);
-  } else {
-    g_set_error(err, OPENSLIDE_ERROR, OPENSLIDE_ERROR_FAILED,
-                "Cannot set TIFF directory");
-  }
+  bool success = _openslide_grid_paint_region(l->grid, cr, tiff,
+                                              x / l->base.downsample,
+                                              y / l->base.downsample,
+                                              level, w, h,
+                                              err);
   _openslide_tiffcache_put(data->tc, tiff);
 
   return success;
@@ -265,6 +275,30 @@ static const struct _openslide_ops ventana_ops = {
   .destroy = destroy,
 };
 
+static xmlNode *get_initial_xml_iscan(xmlDoc *doc, GError **err) {
+  xmlNode *root = xmlDocGetRootElement(doc);
+  if (!xmlStrcmp(root->name, BAD_CAST INITIAL_XML_ISCAN)) {
+    // /iScan
+    return root;
+
+  } else if (!xmlStrcmp(root->name, BAD_CAST INITIAL_XML_ALT_ROOT)) {
+    for (xmlNode *node = root->children; node; node = node->next) {
+      if (!xmlStrcmp(node->name, BAD_CAST INITIAL_XML_ISCAN)) {
+        // /Metadata/iScan, found in some slides
+        return node;
+      }
+    }
+    g_set_error(err, OPENSLIDE_ERROR, OPENSLIDE_ERROR_FAILED,
+                "Couldn't find iScan element in initial XML");
+    return false;
+
+  } else {
+    g_set_error(err, OPENSLIDE_ERROR, OPENSLIDE_ERROR_FAILED,
+                "Unrecognized root element in initial XML");
+    return false;
+  }
+}
+
 static bool ventana_detect(const char *filename G_GNUC_UNUSED,
                            struct _openslide_tifflike *tl,
                            GError **err) {
@@ -275,34 +309,17 @@ static bool ventana_detect(const char *filename G_GNUC_UNUSED,
     return false;
   }
 
-  // find the TIFF directory for level 0
-  int64_t dir_count = _openslide_tifflike_get_directory_count(tl);
-  int64_t dir;
-  for (dir = 0; dir < dir_count; dir++) {
-    const char *image_desc =
-      _openslide_tifflike_get_buffer(tl, dir, TIFFTAG_IMAGEDESCRIPTION, NULL);
-    if (image_desc && strstr(image_desc, LEVEL0_IMAGEDESCRIPTION_MAGIC)) {
-      // found it
-      break;
-    }
-  }
-  if (dir == dir_count) {
-    g_set_error(err, OPENSLIDE_ERROR, OPENSLIDE_ERROR_FAILED,
-                "Not a Ventana file");
-    return false;
-  }
-
   // read XMLPacket
-  const char *xml = _openslide_tifflike_get_buffer(tl, dir, TIFFTAG_XMLPACKET,
+  const char *xml = _openslide_tifflike_get_buffer(tl, 0, TIFFTAG_XMLPACKET,
                                                    err);
   if (!xml) {
     return false;
   }
 
   // check for plausible XML string before parsing
-  if (!strstr(xml, LEVEL0_XML_MAGIC)) {
+  if (!strstr(xml, INITIAL_XML_ISCAN)) {
     g_set_error(err, OPENSLIDE_ERROR, OPENSLIDE_ERROR_FAILED,
-                "%s not in XMLPacket", LEVEL0_XML_MAGIC);
+                "%s not in XMLPacket", INITIAL_XML_ISCAN);
     return false;
   }
 
@@ -312,11 +329,8 @@ static bool ventana_detect(const char *filename G_GNUC_UNUSED,
     return false;
   }
 
-  // check root tag name
-  xmlNode *root = xmlDocGetRootElement(doc);
-  if (xmlStrcmp(root->name, BAD_CAST LEVEL0_XML_ROOT)) {
-    g_set_error(err, OPENSLIDE_ERROR, OPENSLIDE_ERROR_FAILED,
-                "Root tag not %s", LEVEL0_XML_ROOT);
+  // check for iScan element
+  if (!get_initial_xml_iscan(doc, err)) {
     xmlFreeDoc(doc);
     return false;
   }
@@ -325,20 +339,20 @@ static bool ventana_detect(const char *filename G_GNUC_UNUSED,
   return true;
 }
 
-static void slide_info_free(struct slide_info *slide) {
-  if (!slide) {
+static void bif_free(struct bif *bif) {
+  if (!bif) {
     return;
   }
-  for (int32_t i = 0; i < slide->num_areas; i++) {
-    struct area *area = slide->areas[i];
+  for (int32_t i = 0; i < bif->num_areas; i++) {
+    struct area *area = bif->areas[i];
     for (int64_t j = 0; j < area->tile_count; j++) {
       g_slice_free(struct tile, area->tiles[j]);
     }
     g_free(area->tiles);
     g_slice_free(struct area, area);
   }
-  g_free(slide->areas);
-  g_slice_free(struct slide_info, slide);
+  g_free(bif->areas);
+  g_slice_free(struct bif, bif);
 }
 
 static int width_compare(gconstpointer a, gconstpointer b) {
@@ -354,23 +368,17 @@ static int width_compare(gconstpointer a, gconstpointer b) {
   }
 }
 
-static bool process_iscan_metadata(openslide_t *osr, xmlXPathContext *ctx,
-                                   GError **err) {
-  // get node
-  xmlNode *iscan =
-    _openslide_xml_xpath_get_node(ctx, "/EncodeInfo/SlideInfo/iScan");
-  if (!iscan) {
-    g_set_error(err, OPENSLIDE_ERROR, OPENSLIDE_ERROR_FAILED,
-                "Missing or duplicate iScan element");
-    goto FAIL;
+static bool parse_initial_xml(openslide_t *osr, const char *xml,
+                              GError **err) {
+  // parse
+  xmlDoc *doc = _openslide_xml_parse(xml, err);
+  if (!doc) {
+    return false;
   }
 
-  // we don't know how to handle multiple Z layers
-  int64_t z_layers;
-  PARSE_INT_ATTRIBUTE_OR_FAIL(iscan, ATTR_Z_LAYERS, z_layers);
-  if (z_layers != 1) {
-    g_set_error(err, OPENSLIDE_ERROR, OPENSLIDE_ERROR_FAILED,
-                "Slides with multiple Z layers are not supported");
+  // get iScan element
+  xmlNode *iscan = get_initial_xml_iscan(doc, err);
+  if (!iscan) {
     goto FAIL;
   }
 
@@ -393,9 +401,12 @@ static bool process_iscan_metadata(openslide_t *osr, xmlXPathContext *ctx,
   _openslide_duplicate_double_prop(osr, "ventana.ScanRes",
                                    OPENSLIDE_PROPERTY_NAME_MPP_Y);
 
+  // clean up
+  xmlFreeDoc(doc);
   return true;
 
 FAIL:
+  xmlFreeDoc(doc);
   return false;
 }
 
@@ -408,7 +419,7 @@ static bool get_tile_coordinates(const struct area *area,
   PARSE_INT_ATTRIBUTE_OR_FAIL(joint_info, attr_name, tile);
   if (tile < 1 || tile > area->tile_count) {
     g_set_error(err, OPENSLIDE_ERROR, OPENSLIDE_ERROR_FAILED,
-                "Tile number out of bounds: %"G_GINT64_FORMAT, tile);
+                "Tile number out of bounds: %"PRId64, tile);
     goto FAIL;
   }
 
@@ -435,11 +446,10 @@ FAIL:
   return false;
 }
 
-static struct slide_info *parse_level0_xml(openslide_t *osr,
-                                           const char *xml,
-                                           int64_t tiff_tile_width,
-                                           int64_t tiff_tile_height,
-                                           GError **err) {
+static struct bif *parse_level0_xml(const char *xml,
+                                    int64_t tiff_tile_width,
+                                    int64_t tiff_tile_height,
+                                    GError **err) {
   GPtrArray *area_array = g_ptr_array_new();
   xmlXPathContext *ctx = NULL;
   xmlXPathObject *info_result = NULL;
@@ -454,15 +464,11 @@ static struct slide_info *parse_level0_xml(openslide_t *osr,
   // parse
   xmlDoc *doc = _openslide_xml_parse(xml, err);
   if (!doc) {
+    g_prefix_error(err, "Parsing level 0 XML: ");
     goto FAIL;
   }
   ctx = _openslide_xml_xpath_create(doc);
 
-  // read iScan element
-  if (!process_iscan_metadata(osr, ctx, err)) {
-    goto FAIL;
-  }
-
   // query AOI metadata
   info_result =
     _openslide_xml_xpath_eval(ctx, "/EncodeInfo/SlideStitchInfo/ImageInfo");
@@ -501,16 +507,14 @@ static struct slide_info *parse_level0_xml(openslide_t *osr,
     if (tile_width != tiff_tile_width || tile_height != tiff_tile_height) {
       g_set_error(err, OPENSLIDE_ERROR, OPENSLIDE_ERROR_FAILED,
                   "Tile size mismatch: "
-                  "expected %"G_GINT64_FORMAT"x%"G_GINT64_FORMAT", "
-                  "found %"G_GINT64_FORMAT"x%"G_GINT64_FORMAT,
+                  "expected %"PRId64"x%"PRId64", found %"PRId64"x%"PRId64,
                   tiff_tile_width, tiff_tile_height, tile_width, tile_height);
       goto FAIL;
     }
     if (start_col_x % tile_width || start_row_y % tile_height) {
       g_set_error(err, OPENSLIDE_ERROR, OPENSLIDE_ERROR_FAILED,
                   "Area origin not divisible by tile size: "
-                  "%"G_GINT64_FORMAT" %% %"G_GINT64_FORMAT", "
-                  "%"G_GINT64_FORMAT" %% %"G_GINT64_FORMAT,
+                  "%"PRId64" %% %"PRId64", %"PRId64" %% %"PRId64,
                   start_col_x, tile_width, start_row_y, tile_height);
       goto FAIL;
     }
@@ -521,7 +525,16 @@ static struct slide_info *parse_level0_xml(openslide_t *osr,
     PARSE_INT_ATTRIBUTE_OR_FAIL(info, ATTR_NUM_COLS, area->tiles_across);
     PARSE_INT_ATTRIBUTE_OR_FAIL(info, ATTR_NUM_ROWS, area->tiles_down);
 
-    //g_debug("area %d: start %"G_GINT64_FORMAT" %"G_GINT64_FORMAT", count %"G_GINT64_FORMAT" %"G_GINT64_FORMAT, i, area->start_col, area->start_row, area->tiles_across, area->tiles_down);
+    // get position
+    // it seems these are always whole numbers, but they are sometimes
+    // encoded as floating-point values
+    double x, y;
+    PARSE_DOUBLE_ATTRIBUTE_OR_FAIL(info, ATTR_POS_X, x);
+    PARSE_DOUBLE_ATTRIBUTE_OR_FAIL(info, ATTR_POS_Y, y);
+    area->x = x;
+    area->y = y;
+
+    //g_debug("area %d: start %"PRId64" %"PRId64", count %"PRId64" %"PRId64", pos %"PRId64" %"PRId64, i, area->start_col, area->start_row, area->tiles_across, area->tiles_down, area->x, area->y);
 
     // create tile structs
     area->tile_count = area->tiles_across * area->tiles_down;
@@ -558,7 +571,7 @@ static struct slide_info *parse_level0_xml(openslide_t *osr,
       struct joint *joint;
       bool ok;
       bool direction_y = false;
-      //g_debug("%s, tile1 %"G_GINT64_FORMAT" %"G_GINT64_FORMAT", tile2 %"G_GINT64_FORMAT" %"G_GINT64_FORMAT, (char *) direction, tile1_col, tile1_row, tile2_col, tile2_row);
+      //g_debug("%s, tile1 %"PRId64" %"PRId64", tile2 %"PRId64" %"PRId64, (char *) direction, tile1_col, tile1_row, tile2_col, tile2_row);
       if (!xmlStrcmp(direction, BAD_CAST DIRECTION_RIGHT)) {
         // get left joint of right tile
         struct tile *tile =
@@ -581,8 +594,7 @@ static struct slide_info *parse_level0_xml(openslide_t *osr,
       if (!ok) {
         g_set_error(err, OPENSLIDE_ERROR, OPENSLIDE_ERROR_FAILED,
                     "Unexpected tile join: %s, "
-                    "(%"G_GINT64_FORMAT", %"G_GINT64_FORMAT"), "
-                    "(%"G_GINT64_FORMAT", %"G_GINT64_FORMAT")",
+                    "(%"PRId64", %"PRId64"), (%"PRId64", %"PRId64")",
                     (char *) direction, tile1_col, tile1_row,
                     tile2_col, tile2_row);
         xmlFree(direction);
@@ -626,20 +638,40 @@ FAIL:
   }
 
   // create wrapper struct
-  struct slide_info *slide = g_slice_new0(struct slide_info);
-  slide->num_areas = area_array->len;
-  slide->areas = (struct area **) g_ptr_array_free(area_array, false);
-  slide->tile_advance_x = tiff_tile_width + total_offset_x / total_x_weight;
-  slide->tile_advance_y = tiff_tile_height + total_offset_y / total_y_weight;
-  //g_debug("advances: %g %g", slide->tile_advance_x, slide->tile_advance_y);
+  struct bif *bif = g_slice_new0(struct bif);
+  bif->num_areas = area_array->len;
+  bif->areas = (struct area **) g_ptr_array_free(area_array, false);
+  bif->tile_advance_x = tiff_tile_width + total_offset_x / total_x_weight;
+  bif->tile_advance_y = tiff_tile_height + total_offset_y / total_y_weight;
+  //g_debug("advances: %g %g", bif->tile_advance_x, bif->tile_advance_y);
+
+  // Fix area Y coordinates.  The Pos-Y read from the file is the distance
+  // from the bottom of the area to a point below all areas.
+  int64_t *heights = g_new(int64_t, bif->num_areas);
+  // find position of top of slide in coordinate plane of file
+  int64_t top = 0;
+  for (int32_t i = 0; i < bif->num_areas; i++) {
+    struct area *area = bif->areas[i];
+    heights[i] =
+      (area->tiles_down - 1) * bif->tile_advance_y + tiff_tile_height;
+    top = MAX(top, area->y + heights[i]);
+    //g_debug("area %d height %"PRId64, i, heights[i]);
+  }
+  //g_debug("top %"PRId64, top);
+  // convert Y coordinate of each area
+  for (int32_t i = 0; i < bif->num_areas; i++) {
+    struct area *area = bif->areas[i];
+    area->y = top - area->y - heights[i];
+  }
+  g_free(heights);
 
   // free on failure
   if (!success) {
-    slide_info_free(slide);
-    slide = NULL;
+    bif_free(bif);
+    bif = NULL;
   }
 
-  return slide;
+  return bif;
 }
 
 static bool parse_level_info(const char *desc,
@@ -681,8 +713,8 @@ static bool parse_level_info(const char *desc,
   }
 
   // parse magnification
-  *magnification = g_ascii_strtod(magnification_str, &endptr);
-  if (magnification_str[0] == 0 || endptr[0] != 0) {
+  *magnification = _openslide_parse_double(magnification_str);
+  if (isnan(*magnification)) {
     g_set_error(err, OPENSLIDE_ERROR, OPENSLIDE_ERROR_FAILED,
                 "Invalid magnification");
     goto DONE;
@@ -695,28 +727,33 @@ DONE:
   return success;
 }
 
-static struct _openslide_grid *create_grid(openslide_t *osr,
-                                           struct slide_info *slide,
-                                           double downsample,
-                                           int64_t tile_w, int64_t tile_h) {
+static struct _openslide_grid *create_bif_grid(openslide_t *osr,
+                                               struct bif *bif,
+                                               double downsample,
+                                               int64_t tile_w, int64_t tile_h) {
   double subtile_w = tile_w / downsample;
   double subtile_h = tile_h / downsample;
 
   struct _openslide_grid *grid =
     _openslide_grid_create_tilemap(osr,
-                                   slide->tile_advance_x / downsample,
-                                   slide->tile_advance_y / downsample,
-                                   read_subtile, NULL);
+                                   bif->tile_advance_x / downsample,
+                                   bif->tile_advance_y / downsample,
+                                   read_subtile_tilemap, NULL);
 
-  for (int32_t i = 0; i < slide->num_areas; i++) {
-    struct area *area = slide->areas[i];
+  for (int32_t i = 0; i < bif->num_areas; i++) {
+    struct area *area = bif->areas[i];
+    double offset_x =
+      (area->x - area->start_col * bif->tile_advance_x) / downsample;
+    double offset_y =
+      (area->y - area->start_row * bif->tile_advance_y) / downsample;
+    //g_debug("ds %g area %d pos %"PRId64" %"PRId64" offset %g %g", downsample, i, area->x, area->y, offset_x, offset_y);
     for (int64_t row = area->start_row;
          row < area->start_row + area->tiles_down; row++) {
       for (int64_t col = area->start_col;
            col < area->start_col + area->tiles_across; col++) {
         _openslide_grid_tilemap_add_tile(grid,
                                          col, row,
-                                         0, 0,
+                                         offset_x, offset_y,
                                          subtile_w, subtile_h,
                                          NULL);
       }
@@ -726,11 +763,31 @@ static struct _openslide_grid *create_grid(openslide_t *osr,
   return grid;
 }
 
+static void set_region_props(openslide_t *osr, struct bif *bif,
+                             struct level *level0) {
+  for (int32_t i = 0; i < bif->num_areas; i++) {
+    struct area *area = bif->areas[i];
+    g_hash_table_insert(osr->properties,
+                        g_strdup_printf(_OPENSLIDE_PROPERTY_NAME_TEMPLATE_REGION_X, i),
+                        g_strdup_printf("%"PRId64, (int64_t) (bif->tile_advance_x * area->start_col)));
+    g_hash_table_insert(osr->properties,
+                        g_strdup_printf(_OPENSLIDE_PROPERTY_NAME_TEMPLATE_REGION_Y, i),
+                        g_strdup_printf("%"PRId64, (int64_t) (bif->tile_advance_y * area->start_row)));
+    g_hash_table_insert(osr->properties,
+                        g_strdup_printf(_OPENSLIDE_PROPERTY_NAME_TEMPLATE_REGION_WIDTH, i),
+                        g_strdup_printf("%"PRId64, (int64_t) ceil(bif->tile_advance_x * (area->tiles_across - 1) + level0->tiffl.tile_w)));
+    g_hash_table_insert(osr->properties,
+                        g_strdup_printf(_OPENSLIDE_PROPERTY_NAME_TEMPLATE_REGION_HEIGHT, i),
+                        g_strdup_printf("%"PRId64, (int64_t) ceil(bif->tile_advance_y * (area->tiles_down - 1) + level0->tiffl.tile_h)));
+  }
+}
+
 static bool ventana_open(openslide_t *osr, const char *filename,
                          struct _openslide_tifflike *tl,
                          struct _openslide_hash *quickhash1, GError **err) {
   GPtrArray *level_array = g_ptr_array_new();
-  struct slide_info *slide = NULL;
+  struct bif *bif = NULL;
+  GError *tmp_err = NULL;
 
   // open TIFF
   struct _openslide_tiffcache *tc = _openslide_tiffcache_create(filename);
@@ -739,6 +796,16 @@ static bool ventana_open(openslide_t *osr, const char *filename,
     goto FAIL;
   }
 
+  // parse initial XML
+  const char *xml = _openslide_tifflike_get_buffer(tl, 0, TIFFTAG_XMLPACKET,
+                                                   err);
+  if (!xml) {
+    goto FAIL;
+  }
+  if (!parse_initial_xml(osr, xml, err)) {
+    goto FAIL;
+  }
+
   // walk directories
   int64_t next_level = 0;
   double prev_magnification = INFINITY;
@@ -765,13 +832,12 @@ static bool ventana_open(openslide_t *osr, const char *filename,
       // verify that levels and magnifications are properly ordered
       if (level != next_level++) {
         g_set_error(err, OPENSLIDE_ERROR, OPENSLIDE_ERROR_FAILED,
-                    "Unexpected encounter with level %"G_GINT64_FORMAT, level);
+                    "Unexpected encounter with level %"PRId64, level);
         goto FAIL;
       }
       if (magnification >= prev_magnification) {
         g_set_error(err, OPENSLIDE_ERROR, OPENSLIDE_ERROR_FAILED,
-                    "Unexpected magnification in level %"G_GINT64_FORMAT,
-                    level);
+                    "Unexpected magnification in level %"PRId64, level);
         goto FAIL;
       }
       prev_magnification = magnification;
@@ -784,20 +850,26 @@ static bool ventana_open(openslide_t *osr, const char *filename,
 
       // if first level, parse tile info
       if (level == 0) {
-        // get tile size
-        struct _openslide_tiff_level tiffl;
-        if (!_openslide_tiff_level_init(tiff, dir, NULL, &tiffl, err)) {
-          goto FAIL;
-        }
         // get XML
-        const char *xml =
-          _openslide_tifflike_get_buffer(tl, dir, TIFFTAG_XMLPACKET, err);
-        if (!xml) {
-          goto FAIL;
-        }
-        // parse
-        slide = parse_level0_xml(osr, xml, tiffl.tile_w, tiffl.tile_h, err);
-        if (!slide) {
+        xml = _openslide_tifflike_get_buffer(tl, dir, TIFFTAG_XMLPACKET,
+                                             &tmp_err);
+        if (xml) {
+          // get tile size
+          struct _openslide_tiff_level tiffl;
+          if (!_openslide_tiff_level_init(tiff, dir, NULL, &tiffl, err)) {
+            goto FAIL;
+          }
+          // parse
+          bif = parse_level0_xml(xml, tiffl.tile_w, tiffl.tile_h, err);
+          if (!bif) {
+            goto FAIL;
+          }
+        } else if (g_error_matches(tmp_err, OPENSLIDE_ERROR,
+                                   OPENSLIDE_ERROR_NO_VALUE)) {
+          // Ventana TIFF (no AOIs or overlaps)
+          g_clear_error(&tmp_err);
+        } else {
+          g_propagate_error(err, tmp_err);
           goto FAIL;
         }
       }
@@ -826,7 +898,7 @@ static bool ventana_open(openslide_t *osr, const char *filename,
       struct level *l = g_slice_new0(struct level);
       struct _openslide_tiff_level *tiffl = &l->tiffl;
       if (!_openslide_tiff_level_init(tiff, dir,
-                                      NULL, tiffl,
+                                      &l->base, tiffl,
                                       err)) {
         g_slice_free(struct level, l);
         goto FAIL;
@@ -836,16 +908,30 @@ static bool ventana_open(openslide_t *osr, const char *filename,
         level0 = level_array->pdata[0];
       }
       l->base.downsample = downsample;
-      l->grid = create_grid(osr, slide,
-                            downsample,
-                            tiffl->tile_w, tiffl->tile_h);
-      // the format doesn't seem to record the level size, so make it
-      // large enough for all the pixels
-      double x, y, w, h;
-      _openslide_grid_get_bounds(l->grid, &x, &y, &w, &h);
-      l->base.w = ceil(x + w);
-      l->base.h = ceil(y + h);
-      //g_debug("level %"G_GINT64_FORMAT": magnification %g, downsample %g, size %"G_GINT64_FORMAT" %"G_GINT64_FORMAT, level, magnification, downsample, l->base.w, l->base.h);
+      if (bif) {
+        l->grid = create_bif_grid(osr, bif,
+                                  downsample,
+                                  tiffl->tile_w, tiffl->tile_h);
+        l->subtiles_per_tile = downsample;
+        // the format doesn't seem to record the level size, so make it
+        // large enough for all the pixels
+        double x, y, w, h;
+        _openslide_grid_get_bounds(l->grid, &x, &y, &w, &h);
+        l->base.w = ceil(x + w);
+        l->base.h = ceil(y + h);
+        // clear tile size hints set by _openslide_tiff_level_init()
+        l->base.tile_w = 0;
+        l->base.tile_h = 0;
+      } else {
+        l->grid = _openslide_grid_create_simple(osr,
+                                                tiffl->tiles_across,
+                                                tiffl->tiles_down,
+                                                tiffl->tile_w,
+                                                tiffl->tile_h,
+                                                read_subtile);
+        l->subtiles_per_tile = 1;
+      }
+      //g_debug("level %"PRId64": magnification %g, downsample %g, size %"PRId64" %"PRId64, level, magnification, downsample, l->base.w, l->base.h);
 
       // add to array
       g_ptr_array_add(level_array, l);
@@ -860,11 +946,11 @@ static bool ventana_open(openslide_t *osr, const char *filename,
         goto FAIL;
       }
 
-    } else if (!strcmp(image_desc, MACRO_DESCRIPTION)) {
+    } else if (!strcmp(image_desc, MACRO_DESCRIPTION) ||
+               !strcmp(image_desc, MACRO_DESCRIPTION2)) {
       // macro image
       if (!_openslide_tiff_add_associated_image(osr, "macro", tc, dir,
                                                 err)) {
-	g_prefix_error(err, "Can't read macro image: ");
 	goto FAIL;
       }
 
@@ -872,23 +958,33 @@ static bool ventana_open(openslide_t *osr, const char *filename,
       // thumbnail image
       if (!_openslide_tiff_add_associated_image(osr, "thumbnail", tc, dir,
                                                 err)) {
-	g_prefix_error(err, "Can't read thumbnail image: ");
 	goto FAIL;
       }
     }
   } while (TIFFReadDirectory(tiff));
 
-  // free slide info
-  slide_info_free(slide);
-  slide = NULL;
-
   // sort tiled levels
   g_ptr_array_sort(level_array, width_compare);
 
-  // set hash and properties
-  g_assert(level_array->len > 0);
-  struct level *top_level = level_array->pdata[level_array->len - 1];
+  // get level 0
+  if (level_array->len == 0) {
+    g_set_error(err, OPENSLIDE_ERROR, OPENSLIDE_ERROR_FAILED,
+                "No pyramid levels in slide");
+    goto FAIL;
+  }
   struct level *level0 = level_array->pdata[0];
+
+  // set region properties
+  if (bif) {
+    set_region_props(osr, bif, level0);
+  }
+
+  // free bif info
+  bif_free(bif);
+  bif = NULL;
+
+  // set hash and TIFF properties
+  struct level *top_level = level_array->pdata[level_array->len - 1];
   if (!_openslide_tifflike_init_properties_and_hash(osr, tl, quickhash1,
                                                     top_level->tiffl.dir,
                                                     level0->tiffl.dir,
@@ -920,8 +1016,8 @@ static bool ventana_open(openslide_t *osr, const char *filename,
   return true;
 
 FAIL:
-  // free slide info
-  slide_info_free(slide);
+  // free bif info
+  bif_free(bif);
   // free the level array
   if (level_array) {
     for (uint32_t n = 0; n < level_array->len; n++) {
